/* cobrain by Brian Tiffin, April 2013, via pbrain.c */

/* pbrain interpreter in old-style C 
   This is an interpreter for Paul M. Parks's pbrain programming language,
       a variant of Urban Mueller's programming language.
   Do anything you want with this program.
   I welcome bug reports and feature requests.
   Daniel B Cristofani (http://www.hevanet.com/cristofd/)

   20130327, extended and put into service with OpenCOBOL as

   cobrain

   btiffin (https://sourceforge.net/users/btiffin)
   Tectonics: cobc -x cobrainrun.cob subcranial.cob cobrain.c -lreadline
              cobc -x cobrainmain.c subcranial.cob -lreadline
              cobc -x callcobrain.cob cobrain.c
              01 brain-cells.
                 05 brain-cell usage binary-c-long occurs 65536 times.
              CALL "cobrain" USING brain-cells
                  BY CONTENT "file pathname" & x"00"
               or
                  BY CONTENT "{cobrain source text} 99 . 10 ." & x"00"
                  RETURNING cell
              END-CALL
              DISPLAY brain-cell(cell) END-DISPLAY
*/

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <termios.h>
#include <libcob.h>
#include <sys/stat.h>

#define SIZE 65536
#define NUMBITS sizeof(long)*8
#define CA(x) case x: fprintf(stderr, "Error: " 

/* cobrain, a single character read intrepret programming tool */
void (*vectors[256])(int v);

/* colon definitions get passed source-tape offset */
int offsets[256];

/* cobrain repl has an init bloc */
int once = 1;
int bail = 0;
int retcode = 0;

long a[SIZE], memreg;
long *deref;

int s[SIZE], sp, t[SIZE], p, q, length, c, tmp, bytes, scale, tracer, tracing;

char code[SIZE], bin[NUMBITS+1];
char *f, *adr;

FILE *input;

static struct termios oldt;

char module_name[256], *d=module_name;
int (*ocmodule)(int);
void (*ocmodule_minus_1)(int);
void (*ocmodule_minus)(void);
int (*ocmodule_0)(void);
int (*ocmodule_1)(int);
int (*ocmodule_2)(int, int);
int (*ocmodule_3)(int, int, int);
int (*ocmodule_4)(int, int, int, int);
int (*ocmodule_5)(int, int, int, int, int);
int (*ocmodule_6)(int, int, int, int, int, int);

const char *long_to_binary(unsigned long x) {
    unsigned long i=1UL << NUMBITS-1;
    char *d=bin;
    while (i>>=1) *d++=(i & x)?'1':'0';
    *d++ = '\0';
    return bin;
}

int e(int i){
    switch(i){
        CA(2) "call to undefined procedure (%hu) with %d at %d of %s", a[p], p, q, f); break;
        CA(3) "pointer too far %s at %d of %s", p>0?"right":"left", q, f); break;
        CA(4) "unmatched '[' at byte %d of %s", s[sp], f); break;
        CA(5) "unmatched ']' at byte %d of %s", q, f); break;
        CA(6) "unmatched '(' at byte %d of %s", s[sp], f); break;
        CA(7) "unmatched ')' at byte %d of %s", q, f); break;
        CA(8) "can't open %s", f); break;
        CA(9) "unmatched '{' at byte %d of %s", s[sp], f); break;
        CA(10) "unmatched '}' at byte %d of %s", q, f); break;
        CA(11) "unmatched '\"' at byte %d of %s", q, f); break;
        CA(12) "divide by zero at byte %d of %s", q, f); break;
    }
    printf(".\n");
    scale=0;
    sp=0;
    bail=1;
    return(i);
}


void restore_terminal_settings(void)
{
    tcsetattr(0, TCSANOW, &oldt);  /* Apply saved settings */
}

void disable_waiting_for_enter(void)
{
    struct termios newt;

    /* Make terminal read 1 char at a time */
    tcgetattr(0, &oldt);  /* Save terminal settings */
    newt = oldt;  /* Init new settings */
    newt.c_lflag &= ~(ICANON | ECHO);  /* Change settings */
    tcsetattr(0, TCSANOW, &newt);  /* Apply settings */
    atexit(restore_terminal_settings); /* Make sure settings will be restored when program ends  */
}

void opencobol_resolver(void) {
    d=module_name;
    while ((tmp=(code[++q]!='\''))&&q<length) *d++=code[q];
    if (tmp) {e(10); goto error;}
    *d++=0;
    /* OpenCOBOL is just grand */
    //cob_init(0, NULL);  C cobrainmain.c will need the init
    ocmodule=cob_resolve(module_name);
    ocmodule_minus_1 = ocmodule_minus = ocmodule_0 = ocmodule_1 = ocmodule_2 
     = ocmodule_3 = ocmodule_4 = ocmodule_5 = ocmodule_6 = ocmodule;
    if (ocmodule) {
        switch (a[p]) {
            case -2: p--; ocmodule_minus_1(a[p]); break;
            case -1: ocmodule_minus(); break;
            case 0: a[p] = ocmodule_0(); break;
            case 1: if (p<1) {e(3); goto error;} p--; a[p] = ocmodule_1(a[p]); break;
            case 2: if (p<2) {e(3); goto error;} p -= 2; a[p] = ocmodule_2(a[p], a[p+1]); break;
            case 3: if (p<3) {e(3); goto error;} p -= 3; a[p] = ocmodule_3(a[p], a[p+1], a[p+2]); break;
            case 4: if (p<4) {e(3); goto error;} p -= 4; a[p] = ocmodule_4(a[p], a[p+1], a[p+2], a[p+3]); break;
            case 5: if (p<5) {e(3); goto error;} p -= 5; a[p] = ocmodule_5(a[p], a[p+1], a[p+2], a[p+3], a[p+4]); break;
            case 6: if (p<6) {e(3); goto error;} p -= 6; a[p] = ocmodule_6(a[p], a[p+1], a[p+2], a[p+3], a[p+4], a[p+5]); break;
            default:  if (p<1) {e(3); goto error;}
                p--; a[p] = ocmodule_1(a[p]); break;
        }
    }
    //cob_tidy();         C cobrainmain.c will need the tidy
    goto ok;

error:
    exit(3);

ok: ;
}

int opencobol(int val) {
    char module_name[256], *d=module_name;
    int (*ocmodule)(int), r;
    while ((tmp=(code[++q]!='"'))&&q<length) *d++=code[q];
    if (tmp) {e(10); return(0);}
    *d++=0;
    /* OpenCOBOL is just grand */
    //cob_init(0, NULL);  C cobrainmain.c will need the init
    ocmodule=cob_resolve(module_name);
    r=(ocmodule!=NULL)?ocmodule(val):0;
    //cob_tidy();         C cobrainmain.c will need the tidy
    return r;
}

/* the cogbrain interpreter */
int cogbrain(char *args) {


struct stat st;

int v;
unsigned char w;
static void *vectors[] = {
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    /* ! store from register */
    &&implement_33,
    &&implement_34,
    &&implement_35,
    &&implement_36,
    &&implement_37,
    &&implement_38,
    &&implement_39,
    
    &&implement_40,
    &&implement_41,
    &&implement_42,
    &&implement_43,
    &&implement_44,
    &&implement_45,
    &&implement_46,
    &&implement_47,
    &&implement_48,
    &&implement_49,

    &&implement_50,
    &&implement_51,
    &&implement_52,
    &&implement_53,
    &&implement_54,
    &&implement_55,
    &&implement_56,
    &&implement_57,
    &&implement_58,
    &&implement_59,

    &&implement_60,
    &&implement_61,
    &&implement_62,
    &&implement_63,
    &&implement_64,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    &&implement_comment,
    &&implement_91,
    &&implement_92,
    &&implement_93,
    &&implement_94,
    &&implement_95,
    &&implement_96,
    &&implement_comment,
    &&implement_98,
    &&implement_99,

    &&implement_100,
    &&implement_comment,
    &&implement_comment,
    &&implement_103,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_108,
    &&implement_109,

    &&implement_110,
    &&implement_111,
    &&implement_112,
    &&implement_113,
    &&implement_114,
    &&implement_115,
    &&implement_116,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_123,
    &&implement_124,
    &&implement_125,
    &&implement_126,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,

    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment,
    &&implement_comment};

goto start;

/* no-op */
implement_comment:
    scale=0;
    goto next;

/* @ fetch */
implement_33:
    a[p]=memreg;
    scale=0;
    goto next;

/* " lay down string */
implement_34:
    a[++p] = 0;   
    a[++p] = 0;   
    bytes = 0;
    adr = (char *)&a[p];
    while ((tmp=(code[++q] != '"')) && q<length) {
        *adr++ = code[q];
        if (++bytes > sizeof(long)) {bytes = 0; if (code[q+1] != '"') p++;}
    }
    a[++p] = 0;
    if (tmp) {e(11); goto error;}
    scale=0;
    goto next;

/* # output formatted number */
implement_35:
    printf("%020ld ", a[p]);
    scale = 0;
    goto next;

/* $ fetch address of current */
implement_36:
    memreg=(long)&a[p];
    scale = 0;
    goto next;

/* % mod */
implement_37:
    if (p<1) {e(3); goto error;}
    a[p-1] %= a[p--];
    scale = 0;
    goto next;

/* & and */
implement_38:
    if (p<1) {e(3); goto error;}
    a[p-1] &= a[p--];
    scale = 0;
    goto next;

/* ' call module with controlled stack frames*/
implement_39:
    opencobol_resolver();
    scale=0;
    goto next;

/* enter on true ( */
implement_40:
    if (!a[p]) q=t[q];
    scale = 0;
    goto next;

/* close ) is a marker to skip to, but does nothing */
implement_41:
    scale = 0;
    goto next;

/* * multiply */
implement_42:
    if (p<1) {e(3); goto error;}
    a[p-1] *= a[p--];
    scale = 0;
    goto next;

/* + plus */
implement_43:
    if (p<1) {e(3); goto error;}
    a[p-1] += a[p--];
    scale = 0;
    goto next;

/* , charater input */
implement_44:
     if ((c = getchar()) != EOF) {
         a[p] = c == '\n' ? 10 : c;
     } else {
        printf("eof %d ", c);
        a[p]=0;
     }
     scale=0;
    goto next;

/* - minus */    
implement_45:
    if (p<1) {e(3); goto error;}
    a[p-1] -= a[p--];
    scale = 0;
    goto next;

/* . character output */
implement_46:
    putchar(a[p] == 10 ? '\n' : a[p]);
    scale = 0;
    goto next;

/* / integer divide with remainder */
implement_47:
    if (p<1) {e(3); goto error;}
    if (a[p]) {
        tmp=a[p-1];
        a[p-1] /= a[p];
        a[p]=tmp - a[p] * a[p-1];
    }
    scale = 0;
    goto next;

/* digits */
implement_digit:
    if (!scale) p++;
    a[p] *= (scale++) ? 10 : 0;
    a[p] += v;
    goto next;

/* numbers starting with zero, do not advance p, but otherwise scale in */
implement_48:
    a[p] *= (scale++) ? 10 : 0;
    goto next;
implement_49:
    v=1;
    goto implement_digit;
implement_50:
    v=2;
    goto implement_digit;
implement_51:
    v=3;
    goto implement_digit;
implement_52:
    v=4;
    goto implement_digit;
implement_53:
    v=5;
    goto implement_digit;
implement_54:
    v=6;
    goto implement_digit;
implement_55:
    v=7;
    goto implement_digit;
implement_56:
    v=8;
    goto implement_digit;
implement_57:
    v=9;
    goto implement_digit;

/* colon definitions */
implement_colon: 
    s[sp++]=q;
    q = offsets[w];
    scale = 0;
    goto next;

/* colon, define next character, push q, pass execution offset to colon */
implement_58:
    vectors[code[q+1]] = &&implement_colon;
    offsets[code[q+1]] = q+1;
    q = t[q];
    scale = 0;
    goto next;

/* semi colon, pop source tape pointer */
implement_59:
    q=s[--sp];
    scale = 0;
    goto next;

/* Advance and retreat the memory pointer */
implement_60:
    if (--p<0) {e(3); p=0;}
    scale = 0;
    goto next;

/* accept integer to current */
implement_61:
    if (!scanf("%ld", &a[p])) a[p] = 0;
    scale = 0;
    goto next;

/* retreat memory pointer */
implement_62:
    if(++p>=SIZE) {e(3); p=SIZE-1;}
    scale = 0;
    goto next;
    
/* dump memory */
implement_63:
    tmp=p;
    while (tmp-->=0)
        printf("[ %05d is %020ld ]\n", tmp+1, a[tmp+1]);
    scale = 0;
    goto next;


/* fetch to register */
implement_64:
    memreg=a[p];
    scale = 0;
    goto next;


/* looping and conditionals */
/* skip to ] on zero */
implement_91:
    if (!a[p]) q=t[q];
    scale = 0;
    goto next;

/* \ comments */
implement_92:
    while((tmp=(code[++q]!='\\')) && q<length);
    if (tmp) {e(9); goto error;}
    scale = 0;
    goto next;

/* jump back to [ if not zero */
implement_93:
    if (a[p]) q=t[q];
    scale = 0;
    goto next;

/* ^ xor */
implement_94:
    if (p<1) {e(3); goto error;}
    a[p-1] ^= a[p--];
    scale = 0;
    goto next;

/* _ negate */
implement_95:
    a[p] = -a[p];
    scale = 0;
    goto next;

/* ' dereference previous with width in current 1,2,4,8 */
implement_96:
    deref = (long *)a[p-1];
    switch (a[p--]) {
        case 1: a[p] = (char)*deref; break;
        case 2: a[p] = (short)*deref; break;
        case 4: a[p] = (int)*deref; break;
        case 8: a[p] = (long)*deref; break;
        default: a[p] = *deref; break;
    }
    scale = 0;
    goto next;

/* b output binary */
implement_98:
    printf("%64s", long_to_binary(a[p]));
    scale = 0;
    goto next;

/* c compare to zero, -1 0 or 1 left in current */
implement_99:
    a[p] = (a[p] ? ((a[p] < 0) ? -1 : 1) : 0);
    scale = 0;
    goto next;

/* d dup */
implement_100:
    if (p>=SIZE) {e(3); goto error;}
    a[++p]=a[p-1];
    scale = 0;
    goto next;

/* g goback */
implement_103:
    retcode = p+1;
    bail = 1;
    scale = 0;
    goto next;

/* l left shift */
implement_108:
    if (p<1) {e(3); goto error;}
    a[p-1] <<= a[p--];
    scale = 0;
    goto next;

/* m minus */
implement_109:
    a[p]--;
    scale = 0;
    goto next;

/* n output number */
implement_110:
    printf("%d", a[p]);
    scale = 0;
    goto next;

/* o over */
implement_111:
    if (p<1||p>=SIZE) {e(3); goto error;}
    a[++p]=a[p-2];
    scale = 0;
    goto next;

/* p plus */
implement_112:
    a[p]++;
    scale = 0;
    goto next;

/* q quit */
implement_113:
    retcode = -1;
    bail = 1;
    scale = 0;
    goto next;

/* r right shift */
implement_114:
    if (p<1) {e(3); goto error;}
    a[p-1] >>= a[p--];
    scale = 0;
    goto next;

/* s swap */
implement_115:
    if (p<1) {e(3); goto error;}
    tmp=a[p];a[p]=a[p-1];a[p-1]=tmp;
    scale = 0;
    goto next;

/* t trace */
implement_116:
    tracer = (tracer ? 0 : 1);
    scale = 0;
    goto next;

/* { if false */
implement_123:
    if (a[p]) q=t[q];
    scale = 0;
    goto next;

/* | or  */
implement_124:
    if (p<1) {e(3); goto error;}
    a[p-1] |= a[p--];
    scale = 0;
    goto next;

/* } is skip marker and no-op */
implement_125:
    scale = 0;
    goto next;

/* ~ not */
implement_126:
    a[p] = ~a[p];
    scale = 0;
    goto next;

/* starts here */
start:

    //disable_waiting_for_enter();

    if (strncmp("help", args, 4) == 0) {
        /* 1028, report your station */
        /* 1034, send help */
        strncpy(code, "1 'cobrainhelp'", 15);
        f="help";
        length=15;
    } else if (strncmp("file ", args, 5) == 0) {
        args += 5;
        if(!(input = fopen(f=args, "r"))) { e(8); return(0); }
        length = fread(code, 1, SIZE, input);
        fclose(input);
        memset(a, 0, SIZE);
    } else {
        f = "cobrain";
        length=strlen(args);
        strncpy(code, args, SIZE);
    }
    scale=0;
    for (q=0; q<length && !bail; q++){
        switch(code[q]){
            case '(': case '[': case '{': case ':': s[sp++]=q; break;
            case ']': if(!sp--||code[t[t[s[sp]]=q]=s[sp]]!='[') {e(5); return(0);} break;
            case ')': if(!sp--||code[t[t[s[sp]]=q]=s[sp]]!='(') {e(5); return(0);} break;
            case '}': if(!sp--||code[t[t[s[sp]]=q]=s[sp]]!='{') {e(5); return(0);} break;
            case ';': if(!sp--||code[s[sp]]!=':') {e(7); return(0);} t[s[sp]]=q; break;
        }
    }
    if (sp) e(code[s[--sp]]=='['?4:6);

    /* COBOL is an ordinal language, first is 1 */
    retcode = p+1;

    for (q=0; q<length && !bail; q++) {
        if (tracer && !isspace(code[q])) {
            tmp=isgraph(code[q])?code[q]:'.';
            fprintf(stderr, "[ %c at %06d, %05d was %020ld, ", (char)tmp, q, p, a[p]);
            tracing=1;
        }

        w = code[q];
        goto *vectors[code[q]];

next: 
        if (tracing && !isspace(code[q])) {
            fprintf(stderr, " after %05d is %020ld ]\n", p, a[p]);
            tracing=0;
        }
    }

    return(retcode);

error:
    exit(retcode);
}
